"""
Tor Browser Launcher
https://gitlab.torproject.org/tpo/applications/torbrowser-launcher/

Copyright (c) 2013-2023 Micah Lee <micah@micahflee.com>
Copyright (c) 2024 Tor Project

Permission is hereby granted, free of charge, to any person
obtaining a copy of this software and associated documentation
files (the "Software"), to deal in the Software without
restriction, including without limitation the rights to use,
copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following
conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.
"""

import os
import sys
import subprocess
import time
import tarfile
import lzma
import re
import requests
import gpg
import shutil
import xml.etree.ElementTree as ET
from packaging import version

from PyQt5 import QtCore, QtWidgets, QtGui


class TryStableException(Exception):
    pass


class TryDefaultMirrorException(Exception):
    pass


class TryForcingEnglishException(Exception):
    pass


class DownloadErrorException(Exception):
    pass


class Launcher(QtWidgets.QMainWindow):
    """
    Launcher window.
    """

    def __init__(self, common, app, url_list):
        super(Launcher, self).__init__()
        self.common = common
        self.app = app

        self.url_list = url_list
        self.force_redownload = False

        # This is the current version of Tor Browser, which should get updated with every release
        self.min_version = "13.0"

        # Init launcher
        self.set_state(None, "", [])
        self.launch_gui = True

        # If Tor Browser is not installed, detect latest version, download, and install
        if not self.common.settings["installed"] or not self.check_min_version():
            # Different message if downloading for the first time, or because your installed version is too low
            download_message = ""
            if not self.common.settings["installed"]:
                download_message = _("Downloading Tor Browser for the first time.")
            elif not self.check_min_version():
                download_message = _(
                    "Your version of Tor Browser is out-of-date. "
                    "Downloading the newest version."
                )

            # Download and install
            print(download_message)
            self.set_state(
                "task",
                download_message,
                [
                    "download_version_check",
                    "set_version",
                    "download_sig",
                    "download_tarball",
                    "verify",
                    "extract",
                    "run",
                ],
            )

            if self.common.settings["download_over_tor"]:
                print(_("Downloading over Tor"))

        else:
            # Tor Browser is already installed, so run
            launch_message = "Launching Tor Browser."
            print(launch_message)
            self.set_state("task", launch_message, ["run"])

        # Build the rest of the UI

        # Set up the window
        self.setWindowTitle(_("Tor Browser"))
        self.setWindowIcon(QtGui.QIcon(self.common.paths["icon_file"]))

        # Label
        self.label = QtWidgets.QLabel()

        # Progress bar
        self.progress_bar = QtWidgets.QProgressBar()
        self.progress_bar.setTextVisible(True)
        self.progress_bar.setMinimum(0)
        self.progress_bar.setMaximum(0)
        self.progress_bar.setValue(0)

        # Buttons
        self.yes_button = QtWidgets.QPushButton()
        self.yes_button.setIcon(
            self.style().standardIcon(QtWidgets.QStyle.SP_DialogApplyButton)
        )
        self.yes_button.clicked.connect(self.yes_clicked)
        self.start_button = QtWidgets.QPushButton(_("Start"))
        self.start_button.setIcon(
            self.style().standardIcon(QtWidgets.QStyle.SP_DialogApplyButton)
        )
        self.start_button.clicked.connect(self.start)
        self.cancel_button = QtWidgets.QPushButton()
        self.cancel_button.setIcon(
            self.style().standardIcon(QtWidgets.QStyle.SP_DialogCancelButton)
        )
        self.cancel_button.clicked.connect(self.close)
        buttons_layout = QtWidgets.QHBoxLayout()
        buttons_layout.addStretch()
        buttons_layout.addWidget(self.yes_button)
        buttons_layout.addWidget(self.start_button)
        buttons_layout.addWidget(self.cancel_button)
        buttons_layout.addStretch()

        # Layout
        layout = QtWidgets.QVBoxLayout()
        layout.addWidget(self.label)
        layout.addWidget(self.progress_bar)
        layout.addLayout(buttons_layout)

        central_widget = QtWidgets.QWidget()
        central_widget.setLayout(layout)
        self.setCentralWidget(central_widget)

        self.update()

    # Set the current state of Tor Browser Launcher
    def set_state(self, gui, message, tasks, autostart=True):
        self.gui = gui
        self.gui_message = message
        self.gui_tasks = tasks
        self.gui_task_i = 0
        self.gui_autostart = autostart

    # Show and hide parts of the UI based on the current state
    def update(self):
        # Hide widgets
        self.progress_bar.hide()
        self.yes_button.hide()
        self.start_button.hide()

        if "error" in self.gui:
            # Label
            self.label.setText(self.gui_message)

            # Yes button
            if self.gui != "error":
                self.yes_button.setText(_("Yes"))
                self.yes_button.show()

            # Exit button
            self.cancel_button.setText(_("Exit"))

        elif self.gui == "task":
            # Label
            self.label.setText(self.gui_message)

            # Progress bar
            self.progress_bar.show()

            # Start button
            if not self.gui_autostart:
                self.start_button.show()

            # Cancel button
            self.cancel_button.setText(_("Cancel"))

        # Resize the window
        self.adjustSize()

        if self.gui_autostart:
            self.start(None)

    # Yes button clicked, based on the state decide what to do
    def yes_clicked(self):
        if self.gui == "error_try_stable":
            self.try_stable()
        elif self.gui == "error_try_default_mirror":
            self.try_default_mirror()
        elif self.gui == "error_try_tor":
            self.try_tor()

    # Start button clicked, begin tasks
    def start(self, widget, data=None):
        # Hide the start button
        self.start_button.hide()

        # Start running tasks
        self.run_task()

    # Run the next task in the task list
    def run_task(self):
        if self.gui_task_i >= len(self.gui_tasks):
            self.close()
            return

        task = self.gui_tasks[self.gui_task_i]

        # Get ready for the next task
        self.gui_task_i += 1

        if task == "download_version_check":
            print(_("Downloading"), self.common.paths["version_check_url"])
            self.download(
                "version check",
                self.common.paths["version_check_url"],
                self.common.paths["version_check_file"],
            )

        if task == "set_version":
            version = self.get_stable_version()
            if version:
                self.common.build_paths(self.get_stable_version())
                print(_("Latest version: {}").format(version))
                self.run_task()
            else:
                self.set_state(
                    "error", _("Error detecting Tor Browser version."), [], False
                )
                self.update()

        elif task == "download_sig":
            print(
                _("Downloading"),
                self.common.paths["sig_url"].format(self.common.settings["mirror"]),
            )
            self.download(
                "signature", self.common.paths["sig_url"], self.common.paths["sig_file"]
            )

        elif task == "download_tarball":
            print(
                _("Downloading"),
                self.common.paths["tarball_url"].format(self.common.settings["mirror"]),
            )
            if not self.force_redownload and os.path.exists(
                self.common.paths["tarball_file"]
            ):
                self.run_task()
            else:
                self.download(
                    "tarball",
                    self.common.paths["tarball_url"],
                    self.common.paths["tarball_file"],
                )

        elif task == "verify":
            print(_("Verifying Signature"))
            self.verify()

        elif task == "extract":
            print(_("Extracting"), self.common.paths["tarball_filename"])
            self.extract()

        elif task == "run":
            print(_("Running"), self.common.paths["tbb"]["start"])
            self.run()

        elif task == "start_over":
            print(_("Starting download over again"))
            self.start_over()

    def download(self, name, url, path):
        # Download from the selected mirror
        mirror_url = url.format(self.common.settings["mirror"]).encode()

        # Initialize the progress bar
        self.progress_bar.setValue(0)
        self.progress_bar.setMaximum(100)
        if self.common.settings["download_over_tor"]:
            self.progress_bar.setFormat(
                _("Downloading") + " {0} ".format(name) + _("(over Tor)") + ", %p%"
            )
        else:
            self.progress_bar.setFormat(_("Downloading") + " {0}, %p%".format(name))

        def progress_update(total_bytes, bytes_so_far):
            percent = float(bytes_so_far) / float(total_bytes)
            amount = float(bytes_so_far)
            units = "bytes"
            for (size, unit) in [(1024 * 1024, "MiB"), (1024, "KiB")]:
                if amount > size:
                    units = unit
                    amount /= float(size)
                    break

            message = _("Downloaded") + (
                " %2.1f%% (%2.1f %s)" % ((percent * 100.0), amount, units)
            )
            if self.common.settings["download_over_tor"]:
                message += " " + _("(over Tor)")

            self.progress_bar.setMaximum(total_bytes)
            self.progress_bar.setValue(bytes_so_far)
            self.progress_bar.setFormat(message)

        def download_complete():
            # Download complete, next task
            self.run_task()

        def download_error(gui, message):
            print(message)
            self.set_state(gui, message, [], False)
            self.update()

        t = DownloadThread(self.common, mirror_url, path)
        t.progress_update.connect(progress_update)
        t.download_complete.connect(download_complete)
        t.download_error.connect(download_error)
        t.start()
        time.sleep(0.2)

    def try_default_mirror(self):
        # change mirror to default and relaunch TBL
        self.common.settings["mirror"] = self.common.default_mirror
        self.common.save_settings()
        subprocess.Popen([self.common.paths["tbl_bin"]])
        self.close()

    def try_tor(self):
        # set download_over_tor to true and relaunch TBL
        self.common.settings["download_over_tor"] = True
        self.common.save_settings()
        subprocess.Popen([self.common.paths["tbl_bin"]])
        self.close()

    def get_stable_version(self):
        tree = ET.parse(self.common.paths["version_check_file"])
        for up in tree.getroot():
            if up.tag == "update" and up.attrib["appVersion"]:
                version = str(up.attrib["appVersion"])

                # make sure the version does not contain directory traversal attempts
                # e.g. "5.5.3", "6.0a", "6.0a-hardened" are valid but "../../../../.." is invalid
                if not re.match(r"^[a-z0-9\.\-]+$", version):
                    return None

                return version
        return None

    def verify(self):
        self.progress_bar.setValue(0)
        self.progress_bar.setMaximum(0)
        self.progress_bar.show()

        self.label.setText(_("Verifying Signature"))

        def success():
            self.run_task()

        def error(message):
            # Make backup of tarball and sig
            backup_tarball_filename = (
                self.common.paths["tarball_file"] + ".verification_failed"
            )
            backup_sig_filename = self.common.paths["sig_file"] + ".verification_failed"
            shutil.copyfile(self.common.paths["tarball_file"], backup_tarball_filename)
            shutil.copyfile(self.common.paths["sig_file"], backup_sig_filename)

            sigerror = (
                "SIGNATURE VERIFICATION FAILED!\n\n"
                "Error Code: {0}\n\n"
                "You might be under attack, there might be a network problem, or you may be missing a "
                "recently added Tor Browser verification key.\n\n"
                "A copy of the Tor Browser files you downloaded have been saved here:\n"
                "{1}\n{2}\n\n"
                "Click Start to refresh the keyring and try again. If the message persists report the above "
                "error code here:\nhttps://gitlab.torproject.org/tpo/applications/torbrowser-launcher/-/issues/"
            )
            sigerror = sigerror.format(
                message, backup_tarball_filename, backup_sig_filename
            )

            self.set_state("task", sigerror, ["start_over"], False)
            self.update()

        t = VerifyThread(self.common)
        t.error.connect(error)
        t.success.connect(success)
        t.start()
        time.sleep(0.2)

    def extract(self):
        self.progress_bar.setValue(0)
        self.progress_bar.setMaximum(0)
        self.progress_bar.show()

        self.label.setText(_("Installing"))

        def success():
            self.run_task()

        def error(message):
            self.set_state(
                "task",
                _(
                    "Tor Browser Launcher doesn't understand the file format of {0}".format(
                        self.common.paths["tarball_file"]
                    )
                ),
                ["start_over"],
                False,
            )
            self.update()

        t = ExtractThread(self.common)
        t.error.connect(error)
        t.success.connect(success)
        t.start()
        time.sleep(0.2)

    def check_min_version(self):
        installed_version = None
        for line in open(self.common.paths["tbb"]["changelog"], "rb").readlines():
            if line.startswith(b"Tor Browser "):
                installed_version = line.split()[2].decode()
                break

        if version.parse(self.min_version) <= version.parse(installed_version):
            return True

        return False

    def run(self):
        # Don't run if it isn't at least the minimum version
        if not self.check_min_version():
            message = _(
                "The version of Tor Browser you have installed is earlier than it should be, which could be a "
                "sign of an attack!"
            )
            print(message)

            Alert(self.common, message)
            return

        # Run Tor Browser
        subprocess.call(
            [self.common.paths["tbb"]["start"]], cwd=self.common.paths["tbb"]["dir_tbb"]
        )
        sys.exit(0)

    # Start over and download TBB again
    def start_over(self):
        self.force_redownload = True  # Overwrite any existing file
        self.label.setText(_("Downloading Tor Browser over again."))
        self.gui_tasks = ["download_tarball", "verify", "extract", "run"]
        self.gui_task_i = 0
        self.start(None)

    def closeEvent(self, event):
        # Clear the download cache
        try:
            os.remove(self.common.paths["version_check_file"])
            os.remove(self.common.paths["sig_file"])
            os.remove(self.common.paths["tarball_file"])
        except:
            pass

        super(Launcher, self).closeEvent(event)


class Alert(QtWidgets.QMessageBox):
    """
    An alert box dialog.
    """

    def __init__(
        self,
        common,
        message,
        icon=QtWidgets.QMessageBox.NoIcon,
        buttons=QtWidgets.QMessageBox.Ok,
        autostart=True,
    ):
        super(Alert, self).__init__(None)

        self.setWindowTitle(_("Tor Browser Launcher"))
        self.setWindowIcon(QtGui.QIcon(common.paths["icon_file"]))
        self.setText(message)
        self.setIcon(icon)
        self.setStandardButtons(buttons)

        if autostart:
            self.exec_()


class DownloadThread(QtCore.QThread):
    """
    Download a file in a separate thread.
    """

    progress_update = QtCore.pyqtSignal(int, int)
    download_complete = QtCore.pyqtSignal()
    download_error = QtCore.pyqtSignal(str, str)

    def __init__(self, common, url, path):
        super(DownloadThread, self).__init__()
        self.common = common
        self.url = url
        self.path = path

    def run(self):
        with open(self.path, "wb") as f:
            try:
                # Start the request
                r = requests.get(
                    self.url,
                    headers={"User-Agent": "torbrowser-launcher"},
                    stream=True,
                    proxies=self.common.proxies(),
                )

                # If status code isn't 200, something went wrong
                if r.status_code != 200:
                    # Should we use the default mirror?
                    if self.common.settings["mirror"] != self.common.default_mirror:
                        message = (
                            _("Download Error:")
                            + " {0}\n\n"
                            + _("You are currently using a non-default mirror")
                            + ":\n{1}\n\n"
                            + _("Would you like to switch back to the default?")
                        ).format(r.status_code, self.common.settings["mirror"])
                        self.download_error.emit("error_try_default_mirror", message)
                    else:
                        message = (_("Download Error:") + " {0}").format(r.status_code)
                        self.download_error.emit("error", message)

                    r.close()
                    return

                # Start streaming the download
                total_bytes = int(r.headers.get("content-length"))
                bytes_so_far = 0
                for data in r.iter_content(chunk_size=4096):
                    bytes_so_far += len(data)
                    f.write(data)
                    self.progress_update.emit(total_bytes, bytes_so_far)

            except requests.exceptions.SSLError:
                message = _(
                    "Invalid SSL certificate for:\n{0}\n\nYou may be under attack."
                ).format(self.url.decode())
                if not self.common.settings["download_over_tor"]:
                    message += "\n\n" + _("Try the download again using Tor?")
                    self.download_error.emit("error_try_tor", message)
                else:
                    self.download_error.emit("error", message)
                return

            except requests.exceptions.ConnectionError:
                # Connection error
                if self.common.settings["download_over_tor"]:
                    message = _(
                        "Error starting download:\n\n{0}\n\nTrying to download over Tor. "
                        "Are you sure Tor is configured correctly and running?"
                    ).format(self.url.decode())
                    self.download_error.emit("error", message)
                else:
                    message = _(
                        "Error starting download:\n\n{0}\n\nAre you connected to the internet?"
                    ).format(self.url.decode())
                    self.download_error.emit("error", message)

                return

        self.download_complete.emit()


class VerifyThread(QtCore.QThread):
    """
    Verify the signature in a separate thread
    """

    success = QtCore.pyqtSignal()
    error = QtCore.pyqtSignal(str)

    def __init__(self, common):
        super(VerifyThread, self).__init__()
        self.common = common

    def run(self):
        def verify(second_try=False):
            with gpg.Context() as c:
                c.set_engine_info(
                    gpg.constants.protocol.OpenPGP,
                    home_dir=self.common.paths["gnupg_homedir"],
                )

                sig = gpg.Data(file=self.common.paths["sig_file"])
                signed = gpg.Data(file=self.common.paths["tarball_file"])

                try:
                    c.verify(signature=sig, signed_data=signed)
                except gpg.errors.BadSignatures as e:
                    if second_try:
                        self.error.emit(str(e))
                    else:
                        raise Exception
                else:
                    self.success.emit()

        try:
            # Try verifying
            verify()
        except:
            # If it fails, refresh the keyring and try again
            self.common.refresh_keyring()
            verify(True)


class ExtractThread(QtCore.QThread):
    """
    Extract the tarball in a separate thread
    """

    success = QtCore.pyqtSignal()
    error = QtCore.pyqtSignal()

    def __init__(self, common):
        super(ExtractThread, self).__init__()
        self.common = common

    def run(self):
        extracted = False
        try:
            if self.common.paths["tarball_file"][-2:] == "xz":
                # if tarball is .tar.xz
                xz = lzma.LZMAFile(self.common.paths["tarball_file"])
                tf = tarfile.open(fileobj=xz)
                tf.extractall(self.common.paths["tbb"]["dir"])
                extracted = True
            else:
                # if tarball is .tar.gz
                if tarfile.is_tarfile(self.common.paths["tarball_file"]):
                    tf = tarfile.open(self.common.paths["tarball_file"])
                    tf.extractall(self.common.paths["tbb"]["dir"])
                    extracted = True
        except:
            pass

        if extracted:
            self.success.emit()
        else:
            self.error.emit()



"""
Tor Browser Launcher
https://gitlab.torproject.org/tpo/applications/torbrowser-launcher/

Copyright (c) 2013-2023 Micah Lee <micah@micahflee.com>
Copyright (c) 2024 Tor Project

Permission is hereby granted, free of charge, to any person
obtaining a copy of this software and associated documentation
files (the "Software"), to deal in the Software without
restriction, including without limitation the rights to use,
copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following
conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.
"""

import os
import sys
import argparse
import signal

from PyQt5 import QtCore, QtWidgets

from .common import Common, SHARE
from .settings import Settings
from .launcher import Launcher


class Application(QtWidgets.QApplication):
    """
    Qt's QApplication class. It has been overridden to support threads.
    """

    def __init__(self):
        self.setAttribute(QtCore.Qt.AA_X11InitThreads, True)
        QtWidgets.QApplication.__init__(self, sys.argv)
        self.installEventFilter(self)


def main():
    # Parse arguments
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--settings",
        action="store_true",
        dest="settings",
        help="Open Tor Browser Launcher settings",
    )
    parser.add_argument("url", nargs="*", help="URL to load")
    args = parser.parse_args()

    settings = bool(args.settings)
    url_list = args.url

    # Set the TORBROWSER_LAUNCHER env variable to make it easier to
    # detect that torbrowser-launcher is being used
    os.environ["TORBROWSER_LAUNCHER"] = "1"

    # Load the version and print the banner
    with open(os.path.join(SHARE, "version")) as buf:
        tor_browser_launcher_version = buf.read().strip()

    print(_("Tor Browser Launcher"))
    print(_("By Micah Lee & Tor Project, licensed under MIT"))
    print(_("version {0}").format(tor_browser_launcher_version))
    print("https://gitlab.torproject.org/tpo/applications/torbrowser-launcher/")

    common = Common(tor_browser_launcher_version)
    app = Application()

    # Open the window
    gui = None

    if settings:
        # Settings mode
        gui = Settings(common, app)
    else:
        # Launcher mode
        gui = Launcher(common, app, url_list)

    # Center the window
    desktop = app.desktop()
    window_size = gui.size()
    gui.move(
        (desktop.width() - window_size.width()) // 2,
        (desktop.height() - window_size.height()) // 2,
    )
    gui.show()

    # Allow ctrl-c to work
    signal.signal(signal.SIGINT, signal.SIG_DFL)

    sys.exit(app.exec_())


if __name__ == "__main__":
    main()



"""
Tor Browser Launcher
https://gitlab.torproject.org/tpo/applications/torbrowser-launcher/

Copyright (c) 2013-2023 Micah Lee <micah@micahflee.com>
Copyright (c) 2024 Tor Project

Permission is hereby granted, free of charge, to any person
obtaining a copy of this software and associated documentation
files (the "Software"), to deal in the Software without
restriction, including without limitation the rights to use,
copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following
conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.
"""

import subprocess
import shutil

from PyQt5 import QtCore, QtWidgets, QtGui


class Settings(QtWidgets.QMainWindow):
    """
    Settings window.
    """

    def __init__(self, common, app):
        super(Settings, self).__init__()

        self.common = common
        self.app = app

        # Set up the window
        self.setWindowTitle(_("Tor Browser Launcher Settings"))
        self.setWindowIcon(QtGui.QIcon(self.common.paths["icon_file"]))

        # Download over system tor
        self.tor_download_checkbox = QtWidgets.QCheckBox(_("Download over system Tor"))
        if self.common.settings["download_over_tor"]:
            self.tor_download_checkbox.setCheckState(QtCore.Qt.Checked)
        else:
            self.tor_download_checkbox.setCheckState(QtCore.Qt.Unchecked)

        # Tor SOCKS address
        tor_addr_label = QtWidgets.QLabel(_("Tor server"))
        self.tor_addr = QtWidgets.QLineEdit()
        self.tor_addr.setText(self.common.settings["tor_socks_address"])
        tor_addr_layout = QtWidgets.QHBoxLayout()
        tor_addr_layout.addWidget(tor_addr_label)
        tor_addr_layout.addWidget(self.tor_addr)

        # Settings layout
        settings_layout = QtWidgets.QVBoxLayout()
        settings_layout.addWidget(self.tor_download_checkbox)
        settings_layout.addLayout(tor_addr_layout)

        # Status
        status_label = QtWidgets.QLabel()
        if self.common.settings["installed"]:
            status_label.setText(_("Status: Installed"))
        else:
            status_label.setText(_("Status: Not Installed"))

        # Install button
        install_button = QtWidgets.QPushButton(_("Install Tor Browser"))
        install_button.setIcon(
            self.style().standardIcon(QtWidgets.QStyle.SP_DialogApplyButton)
        )
        install_button.clicked.connect(self.install)

        # Reinstall buttons
        reinstall_button = QtWidgets.QPushButton(_("Reinstall Tor Browser"))
        reinstall_button.setIcon(
            self.style().standardIcon(QtWidgets.QStyle.SP_DialogApplyButton)
        )
        reinstall_button.clicked.connect(self.reinstall)

        if self.common.settings["installed"]:
            install_button.hide()
            reinstall_button.show()
        else:
            install_button.show()
            reinstall_button.hide()

        # Status layout
        status_layout = QtWidgets.QVBoxLayout()
        status_layout.addWidget(status_label)
        status_layout.addWidget(install_button)
        status_layout.addWidget(reinstall_button)

        # Top layout
        top_layout = QtWidgets.QHBoxLayout()
        top_layout.addLayout(settings_layout)
        top_layout.addLayout(status_layout)

        # Mirror
        mirror_label = QtWidgets.QLabel(_("Mirror"))

        self.mirror = QtWidgets.QComboBox()
        for mirror in self.common.mirrors:
            self.mirror.addItem(mirror)

        if self.common.settings["mirror"] in self.common.mirrors:
            self.mirror.setCurrentIndex(
                self.mirror.findText(self.common.settings["mirror"])
            )
        else:
            self.mirror.setCurrentIndex(0)

        mirror_layout = QtWidgets.QHBoxLayout()
        mirror_layout.addWidget(mirror_label)
        mirror_layout.addWidget(self.mirror)

        # Save & Exit button
        self.save_exit_button = QtWidgets.QPushButton(_("Save && Exit"))
        self.save_exit_button.setIcon(
            self.style().standardIcon(QtWidgets.QStyle.SP_DialogApplyButton)
        )
        self.save_exit_button.clicked.connect(self.save_exit)

        # Cancel button
        self.cancel_button = QtWidgets.QPushButton(_("Cancel"))
        self.cancel_button.setIcon(
            self.style().standardIcon(QtWidgets.QStyle.SP_DialogCancelButton)
        )
        self.cancel_button.clicked.connect(self.close)

        # Buttons layout
        buttons_layout = QtWidgets.QHBoxLayout()
        buttons_layout.addWidget(self.save_exit_button)
        buttons_layout.addWidget(self.cancel_button)

        # Main layout
        layout = QtWidgets.QVBoxLayout()
        layout.addLayout(top_layout)
        layout.addLayout(mirror_layout)
        layout.addLayout(buttons_layout)

        central_widget = QtWidgets.QWidget()
        central_widget.setLayout(layout)
        self.setCentralWidget(central_widget)

    # Install
    def install(self):
        self.save()
        subprocess.Popen([self.common.paths["tbl_bin"]])
        self.close()

    # Reinstall
    def reinstall(self):
        self.save()
        shutil.rmtree(self.common.paths["tbb"]["dir"])
        subprocess.Popen([self.common.paths["tbl_bin"]])
        self.close()

    # Save & Exit
    def save_exit(self):
        self.save()
        self.close()

    # Save settings
    def save(self):
        # Checkbox options
        self.common.settings[
            "download_over_tor"
        ] = self.tor_download_checkbox.isChecked()
        self.common.settings["tor_socks_address"] = self.tor_addr.text()

        # Figure out the selected mirror
        self.common.settings["mirror"] = self.mirror.currentText()

        # Save them
        self.common.save_settings()



"""
Tor Browser Launcher
https://gitlab.torproject.org/tpo/applications/torbrowser-launcher/

Copyright (c) 2013-2021 Micah Lee <micah@micahflee.com>

Permission is hereby granted, free of charge, to any person
obtaining a copy of this software and associated documentation
files (the "Software"), to deal in the Software without
restriction, including without limitation the rights to use,
copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following
conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.
"""

import os
import sys
import platform
import subprocess
import pickle
import json
import re
import gettext
import gpg
import requests

SHARE = os.getenv("TBL_SHARE", sys.prefix + "/share") + "/torbrowser-launcher"

gettext.install("torbrowser-launcher")


class Common(object):
    def __init__(self, tbl_version):
        self.tbl_version = tbl_version

        # initialize the app
        self.architecture = "x86_64" if "64" in platform.architecture()[0] else "i686"
        self.default_mirror = "https://dist.torproject.org/"
        self.build_paths()
        self.torbrowser12_rename_old_tbb()
        for d in self.paths["dirs"]:
            self.mkdir(self.paths["dirs"][d])
        self.load_mirrors()
        self.load_settings()
        self.build_paths()
        self.mkdir(self.paths["download_dir"])
        self.mkdir(self.paths["tbb"]["dir"])
        self.init_gnupg()

    # get value of environment variable, if it is not set return the default value
    @staticmethod
    def get_env(var_name, default_value):
        value = os.getenv(var_name)
        if not value:
            value = default_value
        return value

    # build all relevant paths
    def build_paths(self, tbb_version=None):
        homedir = os.getenv("HOME")
        if not homedir:
            homedir = "/tmp/.torbrowser-" + os.getenv("USER")
            if not os.path.exists(homedir):
                try:
                    os.mkdir(homedir, 0o700)
                except:
                    self.set_gui(
                        "error", _("Error creating {0}").format(homedir), [], False
                    )

        tbb_config = "{0}/torbrowser".format(
            self.get_env("XDG_CONFIG_HOME", "{0}/.config".format(homedir))
        )
        tbb_cache = "{0}/torbrowser".format(
            self.get_env("XDG_CACHE_HOME", "{0}/.cache".format(homedir))
        )
        tbb_local = "{0}/torbrowser".format(
            self.get_env("XDG_DATA_HOME", "{0}/.local/share".format(homedir))
        )
        old_tbb_data = "{0}/.torbrowser".format(homedir)

        if tbb_version:
            # tarball filename
            if self.architecture == "x86_64":
                arch = "linux-x86_64"
            else:
                arch = "linux-i686"

            tarball_filename = "tor-browser-" + arch + "-" + tbb_version + ".tar.xz"

            # tarball
            self.paths["tarball_url"] = (
                "{0}torbrowser/" + tbb_version + "/" + tarball_filename
            )
            self.paths["tarball_file"] = tbb_cache + "/download/" + tarball_filename
            self.paths["tarball_filename"] = tarball_filename

            # sig
            self.paths["sig_url"] = (
                "{0}torbrowser/" + tbb_version + "/" + tarball_filename + ".asc"
            )
            self.paths["sig_file"] = (
                tbb_cache + "/download/" + tarball_filename + ".asc"
            )
            self.paths["sig_filename"] = tarball_filename + ".asc"
        else:
            self.paths = {
                "dirs": {
                    "config": tbb_config,
                    "cache": tbb_cache,
                    "local": tbb_local,
                },
                "old_data_dir": old_tbb_data,
                "tbl_bin": sys.argv[0],
                "icon_file": os.path.join(
                    os.path.dirname(SHARE), "pixmaps/torbrowser.png"
                ),
                "torproject_pem": os.path.join(SHARE, "torproject.pem"),
                "signing_keys": {
                    "tor_browser_developers": os.path.join(
                        SHARE, "tor-browser-developers.asc"
                    ),
                    "wkd_tmp": os.path.join(tbb_cache, "torbrowser.gpg"),
                },
                "mirrors_txt": [
                    os.path.join(SHARE, "mirrors.txt"),
                    tbb_config + "/mirrors.txt",
                ],
                "download_dir": tbb_cache + "/download",
                "gnupg_homedir": tbb_local + "/gnupg_homedir",
                "settings_file": tbb_config + "/settings.json",
                "settings_file_pickle": tbb_config + "/settings",
                "version_check_url": "https://aus1.torproject.org/torbrowser/update_3/release/Linux_x86_64-gcc3/x/ALL",
                "version_check_file": tbb_cache + "/download/release.xml",
                "tbb": {
                    "changelog": tbb_local
                    + "/tbb/"
                    + self.architecture
                    + "/tor-browser/Browser/TorBrowser/Docs/ChangeLog.txt",
                    "dir": tbb_local + "/tbb/" + self.architecture,
                    "dir_tbb": tbb_local + "/tbb/" + self.architecture + "/tor-browser",
                    "start": tbb_local
                    + "/tbb/"
                    + self.architecture
                    + "/tor-browser/start-tor-browser.desktop",
                },
            }

        # Add the expected fingerprint for imported keys:
        tor_browser_developers_fingerprint = "EF6E286DDA85EA2A4BA7DE684E2C6E8793298290"
        self.fingerprints = {
            "tor_browser_developers": tor_browser_developers_fingerprint,
            "wkd_tmp": tor_browser_developers_fingerprint,
        }

    # Tor Browser 12.0 no longer has locales. If an old TBB folder exists with locals, rename it to just tor_browser
    def torbrowser12_rename_old_tbb(self):
        if not os.path.exists(self.paths["tbb"]["dir"]):
            return
        for filename in os.listdir(self.paths["tbb"]["dir"]):
            abs_filename = os.path.join(self.paths["tbb"]["dir"], filename)
            if filename.startswith("tor-browser_") and os.path.isdir(abs_filename):
                os.rename(abs_filename, self.paths["tbb"]["dir_tbb"])
                print(
                    _("Renamed {0} to {1}").format(
                        abs_filename, self.paths["tbb"]["dir_tbb"]
                    )
                )
                break

    # create a directory
    @staticmethod
    def mkdir(path):
        try:
            if not os.path.exists(path):
                os.makedirs(path, 0o700)
                return True
        except:
            print(_("Cannot create directory {0}").format(path))
            return False
        if not os.access(path, os.W_OK):
            print(_("{0} is not writable").format(path))
            return False
        return True

    # if gnupg_homedir isn't set up, set it up
    def init_gnupg(self):
        if not os.path.exists(self.paths["gnupg_homedir"]):
            print(_("Creating GnuPG homedir"), self.paths["gnupg_homedir"])
            self.mkdir(self.paths["gnupg_homedir"])
        self.import_keys()

    def proxies(self):
        # Use tor socks5 proxy, if enabled
        if self.settings["download_over_tor"]:
            socks5_address = "socks5h://{}".format(self.settings["tor_socks_address"])
            return {"https": socks5_address, "http": socks5_address}
        else:
            return None

    def refresh_keyring(self):
        print("Downloading latest Tor Browser signing key...")

        # Fetch key from wkd, as per https://support.torproject.org/tbb/how-to-verify-signature/
        # Sometimes GPG throws errors, so comment this out and download it directly
        # p = subprocess.Popen(
        #     [
        #         "gpg",
        #         "--status-fd",
        #         "2",
        #         "--homedir",
        #         self.paths["gnupg_homedir"],
        #         "--auto-key-locate",
        #         "nodefault,wkd",
        #         "--locate-keys",
        #         "torbrowser@torproject.org",
        #     ],
        #     stderr=subprocess.PIPE,
        # )
        # p.wait()

        # Download the key from WKD directly
        r = requests.get(
            "https://torproject.org/.well-known/openpgpkey/hu/kounek7zrdx745qydx6p59t9mqjpuhdf?l=torbrowser",
            proxies=self.proxies(),
        )
        if r.status_code != 200:
            print(f"Error fetching key, status code = {r.status_code}")
        else:
            with open(self.paths["signing_keys"]["wkd_tmp"], "wb") as f:
                f.write(r.content)

            if self.import_key_and_check_status("wkd_tmp"):
                print("Key imported successfully")
            else:
                print("Key failed to import")

    def import_key_and_check_status(self, key):
        """Import a GnuPG key and check that the operation was successful.
        :param str key: A string specifying the key's filepath from
            ``Common.paths``
        :rtype: bool
        :returns: ``True`` if the key is now within the keyring (or was
            previously and hasn't changed). ``False`` otherwise.
        """
        with gpg.Context() as c:
            c.set_engine_info(
                gpg.constants.protocol.OpenPGP, home_dir=self.paths["gnupg_homedir"]
            )

            impkey = self.paths["signing_keys"][key]
            try:
                c.op_import(gpg.Data(file=impkey))
            except:
                return False
            else:
                result = c.op_import_result()
                if result and self.fingerprints[key] in result.imports[0].fpr:
                    return True
                else:
                    return False

    # import gpg keys
    def import_keys(self):
        """Import all GnuPG keys.
        :rtype: bool
        :returns: ``True`` if all keys were successfully imported; ``False``
            otherwise.
        """
        keys = [
            "tor_browser_developers",
        ]
        all_imports_succeeded = True

        for key in keys:
            imported = self.import_key_and_check_status(key)
            if not imported:
                print(
                    _(
                        "Could not import key with fingerprint: %s."
                        % self.fingerprints[key]
                    )
                )
                all_imports_succeeded = False

        if not all_imports_succeeded:
            print(_("Not all keys were imported successfully!"))

        return all_imports_succeeded

    # load mirrors
    def load_mirrors(self):
        self.mirrors = []
        for srcfile in self.paths["mirrors_txt"]:
            if not os.path.exists(srcfile):
                continue
            for mirror in open(srcfile, "r").readlines():
                if mirror.strip() not in self.mirrors:
                    self.mirrors.append(mirror.strip())

    # load settings
    def load_settings(self):
        default_settings = {
            "tbl_version": self.tbl_version,
            "installed": False,
            "download_over_tor": False,
            "tor_socks_address": "127.0.0.1:9050",
            "mirror": self.default_mirror,
        }

        if os.path.isfile(self.paths["settings_file"]):
            settings = json.load(open(self.paths["settings_file"]))
            resave = False

            # detect installed
            settings["installed"] = os.path.isfile(self.paths["tbb"]["start"])

            # make sure settings file is up-to-date
            for setting in default_settings:
                if setting not in settings:
                    settings[setting] = default_settings[setting]
                    resave = True

            # make sure tor_socks_address doesn't start with 'tcp:'
            if settings["tor_socks_address"].startswith("tcp:"):
                settings["tor_socks_address"] = settings["tor_socks_address"][4:]
                resave = True

            # make sure the version is current
            if settings["tbl_version"] != self.tbl_version:
                settings["tbl_version"] = self.tbl_version
                resave = True

            self.settings = settings
            if resave:
                self.save_settings()

        # if settings file is still using old pickle format, convert to json
        elif os.path.isfile(self.paths["settings_file_pickle"]):
            self.settings = pickle.load(open(self.paths["settings_file_pickle"]))
            self.save_settings()
            os.remove(self.paths["settings_file_pickle"])
            self.load_settings()

        else:
            self.settings = default_settings
            self.save_settings()

    # save settings
    def save_settings(self):
        json.dump(self.settings, open(self.paths["settings_file"], "w"))
        return True
